/* $LAAS: server.c 2008/07/15 11:25:07 mallet $ */

/*
 * Copyright (c) 2008 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 *					Anthony Mallet on Thu Jul  3 2008
 */

#define WINVER WindowsXP
#include <w32api.h>

#include <stdio.h>
#include <getopt.h>
#include <windows.h>
#include <winsock2.h>

#include "EVaRT2.h"
#include "evart-client.h"

#include "internal.h"


/* --- local data ---------------------------------------------------- */

static SOCKET servfd;

static struct {
  evas_msgtype_t type;
  struct sockaddr addr;
  char name[128];
} requests[64];

static int	evas_serverlisten(SOCKET sockfd);
static void	evas_datahandler(sFrameOfData* FrameOfData);
static int	evas_requestnew(evas_msgtype_t type, struct sockaddr *addr,
			const char *name);
static int	evas_requestfree(evas_msgtype_t type, struct sockaddr *addr,
			const char *name);





void Request()
{
    int retval;
    void *pResponse;
    int nBytes;
    char key;

    while (1)
    {
        printf("\n");
        printf("Options(<esc>=Back):\n");
        printf("----------------\n");
        printf("1) LiveMode\n");
        printf("2) Pause\n");
        printf("3) SetOutputName=<Filename>\n");
        printf("4) StartRecording\n");
        printf("5) StopRecording\n");
        printf("6) ResetIDs\n");
        printf("7) GetContextFrameRate\n");
        printf("8) GetUpAxis\n");
        printf("9) GetConversionToMillimeters\n");
        printf("----------------\n");
        printf("--> ");

        key = getch();

        printf("%c\n", key);
        printf("\n");

        if (key == 27)
        {
            return;
        }


        if (key == '1')
        {
            retval = EVaRT2_Request("LiveMode", &pResponse, &nBytes);
            if (retval != RC_Okay) goto NOT_OKAY;
        }
        else
        if (key == '2')
        {
            retval = EVaRT2_Request("PauseMode", &pResponse, &nBytes);
            if (retval != RC_Okay) goto NOT_OKAY;
        }
        else
        if (key == '3')
        {
            char szRequest[256];
            char szFilename[256];

            printf("Filename --> ");
            scanf("%s\n", szFilename);
            sprintf(szRequest, "SetOutputName=%s", szFilename);
            retval = EVaRT2_Request(szRequest, &pResponse, &nBytes);
            if (retval != RC_Okay) goto NOT_OKAY;
        }
        else
        if (key == '4')
        {
            retval = EVaRT2_Request("StartRecording", &pResponse, &nBytes);
            if (retval != RC_Okay) goto NOT_OKAY;
        }
        else
        if (key == '5')
        {
            retval = EVaRT2_Request("StopRecording", &pResponse, &nBytes);
            if (retval != RC_Okay) goto NOT_OKAY;
        }
        else
        if (key == '6')
        {
            retval = EVaRT2_Request("ResetIDs", &pResponse, &nBytes);
            if (retval != RC_Okay) goto NOT_OKAY;
        }
        else
        if (key == '7')
        {
            retval = EVaRT2_Request("GetContextFrameRate", &pResponse, &nBytes);
            if (retval != RC_Okay) goto NOT_OKAY;

            printf("ContextFrameRate = %f\n", *(float*)pResponse);
        }
        else
        if (key == '8')
        {
            retval = EVaRT2_Request("GetUpAxis", &pResponse, &nBytes);
            if (retval != RC_Okay) goto NOT_OKAY;

            printf("UpAxis = %d\n", *(int*)pResponse);
        }
        else
        if (key == '9')
        {
            retval = EVaRT2_Request("GetConversionToMillimeters", &pResponse, &nBytes);
            if (retval != RC_Okay) goto NOT_OKAY;

            printf("ConversionToMillimeters = %f\n", *(float*)pResponse);
        }




NOT_OKAY:
        if (retval == RC_TimeOut)
        {
            printf("TimeOut\n");
        }
        else
        if (retval == RC_Unrecognized)
        {
            printf("Unrecognized\n");
        }
    }

 // Unreachable
    return;
}


/* --- main ---------------------------------------------------------- */

int
main(int argc, char* argv[])
{
  WSADATA supported;
  int s;

  /* options */
  static const char *usage =
    "usage: %s [--host evart_host] [--localhost this_host]\n"
    "\t\t\t[--port port_number]\n";
  static const struct option long_options[] = {
    { "host", 1, NULL, 'h' },
    { "localhost", 1, NULL, 'l' },
    { "port", 1, NULL, 'p' },
    { NULL, 0, NULL, 0 }
  };
  char *host = NULL;
  char *localhost = NULL;
  unsigned short port = 3615;

  while(1) {
    s = getopt_long(argc, argv, "h:l:p:", long_options, NULL);
    if (s == -1) break;

    switch (s) {
      case 'h':	host = optarg; break;
      case 'l': localhost = optarg; break;
      case 'p': port = atoi(optarg); break;

      case '?':
      default:
	printf(usage, argv[0]);
	exit(2);
    }
  }
  if (optind < argc) {
    printf(usage, argv[0]);
    exit(2);
  }

  /* setup sockets */
  s = WSAStartup(MAKEWORD(2, 2), &supported);
  if (s) {
    fprintf(stderr, "could not initialize sockets\n");
    return 2;
  }
  if (LOBYTE(supported.wVersion) != 2 || HIBYTE(supported.wVersion) != 2) {
    fprintf(stderr, "could not find a useable version of winsock.dll\n");
    WSACleanup();
    return 2;
  }

  /* connect to EVaRT */
  s = evas_evartinit(evas_datahandler);
  if (s) return s;

  s = evas_evartconnect(localhost, host);
  if (s) return s;

  /* setup server */
  servfd = evas_servercreate(port);
  if (servfd == INVALID_SOCKET) goto done;

  s = evas_serverlisten(servfd);

done:
    s = EVaRT2_Exit();
    printf("press any key to terminate...");
    s = getch();

    WSACleanup();
    return 0;
}


/* --- evas_servercreate --------------------------------------------- */

SOCKET
evas_servercreate(unsigned short port)
{
  SOCKET sockfd;
  struct sockaddr_in addr;
  int reuse;

  sockfd = socket(AF_INET, SOCK_DGRAM, 0);
  if (sockfd == INVALID_SOCKET) return INVALID_SOCKET;

  /* set reuse addr option */
  reuse = 1;
  if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR,
		 (char *)&reuse, sizeof(reuse))) {
    perror("cannot set reuseaddr on server socket");
    goto error;
  }

  /* bind */
  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr = htonl(INADDR_ANY);
  addr.sin_port = htons(port);
  if (bind(sockfd, (struct sockaddr *)&addr, sizeof(struct sockaddr_in)))
    goto error;

  return sockfd;

error:
  fprintf(stderr, "cannot create socket server\n");
  closesocket(sockfd);
  return INVALID_SOCKET;
}


/* --- evas_serverlisten --------------------------------------------- */

static int
evas_serverlisten(SOCKET sockfd)
{
  char buffer[65536];
  evas_msg_t *msg = (evas_msg_t *)buffer;
  struct sockaddr from;
  int fromlen = sizeof(from);
  int s;

  while(1) {
    s = recvfrom(sockfd, buffer, sizeof(buffer), 0, &from, &fromlen);
    if (s <= 0) continue;

    switch(msg->type) {
      case EVAS_NONE: break;

      case EVAS_MSG:
	printf("* %s\n", msg->cmd);
	if (!strncmp(msg->cmd, "hello", 5)) {
	  msg->type = EVAS_MSG;
	  s = snprintf(msg->cmd, sizeof(msg->cmd), "evart-server");
	  msg->len = (char *)&msg->cmd[s+1] - (char *)msg;
	  s = sendto(sockfd, buffer, msg->len, 0, &from, fromlen);
	}
	break;

      case EVAS_BODY_MARKERS:
	switch(msg->cmd[0]) {
	  case '-':
	    s = evas_requestfree(EVAS_BODY_MARKERS, &from, msg->cmd+1);
	    break;

	  case '+':
	    s = evas_requestnew(EVAS_BODY_MARKERS, &from, msg->cmd+1);
	    break;

	  default:
	    s = evas_requestnew(EVAS_BODY_MARKERS, &from, msg->cmd);
	    break;
	}
	break;

      case EVAS_UNRECOGNIZED_MARKERS:
	switch(msg->cmd[0]) {
	  case '-':
	    s = evas_requestfree(EVAS_UNRECOGNIZED_MARKERS, &from, "");
	    break;

	  case '+':
	  default:
	    s = evas_requestnew(EVAS_UNRECOGNIZED_MARKERS, &from, "");
	    break;
	}
	break;
    }
  }

  return 0;
}


/* --- evas_datahandler ---------------------------------------------- */

static void
evas_datahandler(sFrameOfData *frame)
{
  char buffer[65536];
  evas_msg_t *msg = (evas_msg_t *)buffer;
  int i, j, k;

  printf("%d\r", frame->iFrame);
  for(i=0; i<sizeof(requests)/sizeof(requests[0]); i++)
    switch(requests[i].type) {
      case EVAS_NONE: break;

      case EVAS_BODY_MARKERS:
	for(j=0; j<frame->nBodies; j++) {
	  if (strcmp(frame->BodyData[j].szName, requests[i].name)) continue;

	  msg->len =
	    (char *)msg->body_markers.markers[frame->BodyData[j].nMarkers] -
	    (char *)msg;
	  if (msg->len > sizeof(buffer)) {
	    fprintf(stderr, "too many markers (%d) for body %s\n",
		    frame->BodyData[j].nMarkers,
		    frame->BodyData[j].szName);
	    continue;
	  }
	  msg->type = EVAS_BODY_MARKERS;
	  strncpy(msg->body_markers.name, frame->BodyData[j].szName,
		  sizeof(msg->body_markers.name));
	  msg->body_markers.nmarkers = frame->BodyData[j].nMarkers;
	  for(k=0; k<msg->body_markers.nmarkers; k++) {
	    msg->body_markers.markers[k][0] = frame->BodyData[j].Markers[k][0];
	    msg->body_markers.markers[k][1] = frame->BodyData[j].Markers[k][1];
	    msg->body_markers.markers[k][2] = frame->BodyData[j].Markers[k][2];
	  }

	  sendto(servfd, buffer, msg->len, 0,
		 &requests[i].addr, sizeof(requests[i].addr));
	}
	break;

      case EVAS_UNRECOGNIZED_MARKERS:
	msg->len =
	  (char *)msg->body_markers.markers[frame->nUnidentifiedMarkers] -
	  (char *)msg;
	if (msg->len > sizeof(buffer)) {
	  fprintf(stderr, "too many unidentified markers (%d)\n",
		  frame->nUnidentifiedMarkers);
	  continue;
	}
	msg->type = EVAS_UNRECOGNIZED_MARKERS;
	msg->unrecognized_markers.nmarkers = frame->nUnidentifiedMarkers;
	for(k=0; k<msg->unrecognized_markers.nmarkers; k++) {
	  msg->unrecognized_markers.markers[k][0] = frame->UnidentifiedMarkers[k][0];
	  msg->unrecognized_markers.markers[k][1] = frame->UnidentifiedMarkers[k][1];
	  msg->unrecognized_markers.markers[k][2] = frame->UnidentifiedMarkers[k][2];
	}

	sendto(servfd, buffer, msg->len, 0,
	       &requests[i].addr, sizeof(requests[i].addr));
	break;
    }
}


/* --- evas_requestnew ----------------------------------------------- */

static int
evas_requestnew(evas_msgtype_t type, struct sockaddr *addr, const char *name)
{
  int i;

  for(i=0; i<sizeof(requests)/sizeof(requests[0]); i++)
    if (requests[i].type == EVAS_NONE) {
      requests[i].addr = *addr;
      snprintf(requests[i].name, sizeof(requests[i].name), "%s", name);
      requests[i].type = type;
      return 0;
    }

  return -1;
}


/* --- evas_requestfree ---------------------------------------------- */

static int
evas_requestfree(evas_msgtype_t type, struct sockaddr *addr, const char *name)
{
  int i;

  for(i=0; i<sizeof(requests)/sizeof(requests[0]); i++)
    if (requests[i].type == type) {
      if (memcmp(addr, &requests[i].addr, sizeof(*addr))) continue;
      if (strcmp(name, requests[i].name)) continue;

      requests[i].type = EVAS_NONE;
      return 0;
    }

  return -1;
}
